# ifndef CPPAD_CPPAD_IPOPT_EXAMPLE_ODE_CHECK_HPP
# define CPPAD_CPPAD_IPOPT_EXAMPLE_ODE_CHECK_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

extern bool ode_check(const SizeVector& N, const NumberVector& x);

# endif
