# ifndef CPPAD_CORE_SPARSE_HPP
# define CPPAD_CORE_SPARSE_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

//
# include <cppad/core/for_jac_sparsity.hpp>
# include <cppad/core/rev_jac_sparsity.hpp>
//
# include <cppad/core/for_hes_sparsity.hpp>
# include <cppad/core/rev_hes_sparsity.hpp>
//
# include <cppad/core/for_sparse_jac.hpp>
# include <cppad/core/rev_sparse_jac.hpp>
//
# include <cppad/core/for_sparse_hes.hpp>
# include <cppad/core/rev_sparse_hes.hpp>
//
# include <cppad/core/sparse_jac.hpp>
# include <cppad/core/sparse_hes.hpp>
//
# include <cppad/core/sparse_jacobian.hpp>
# include <cppad/core/sparse_hessian.hpp>
//
# include <cppad/core/subgraph_sparsity.hpp>
# include <cppad/core/subgraph_reverse.hpp>
# include <cppad/core/subgraph_jac_rev.hpp>

# endif
