# ifndef CPPAD_CORE_ARITHMETIC_HPP
# define CPPAD_CORE_ARITHMETIC_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

/*
-------------------------------------------------------------------------------
{xrst_begin Arithmetic}

AD Arithmetic Operators and Compound Assignments
################################################

Contents
********
{xrst_toc_table
   include/cppad/core/unary_plus.hpp
   include/cppad/core/unary_minus.hpp
   include/cppad/core/ad_binary.hpp
   include/cppad/core/compound_assign.hpp
}

{xrst_end Arithmetic}
-------------------------------------------------------------------------------
*/
# include <cppad/core/unary_plus.hpp>
# include <cppad/core/unary_minus.hpp>
# include <cppad/core/ad_binary.hpp>
# include <cppad/core/compound_assign.hpp>

# endif
