# ifndef CPPAD_LOCAL_VAR_OP_VAR_OP_HPP
# define CPPAD_LOCAL_VAR_OP_VAR_OP_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-24 Bradley M. Bell
// ----------------------------------------------------------------------------

// used by the sparse operators
# include <cppad/local/sparse/internal.hpp>

// operations
# include <cppad/core/std_math_11.hpp>
# include <cppad/local/var_op/abs_op.hpp>
# include <cppad/local/var_op/add_op.hpp>
# include <cppad/local/var_op/acos_op.hpp>
# include <cppad/local/var_op/acosh_op.hpp>
# include <cppad/local/var_op/asin_op.hpp>
# include <cppad/local/var_op/asinh_op.hpp>
# include <cppad/local/var_op/atan_op.hpp>
# include <cppad/local/var_op/atanh_op.hpp>
# include <cppad/local/var_op/cexp_op.hpp>
# include <cppad/local/var_op/cos_op.hpp>
# include <cppad/local/var_op/cosh_op.hpp>
# include <cppad/local/var_op/cskip_op.hpp>
# include <cppad/local/var_op/csum_op.hpp>
# include <cppad/local/var_op/dis_op.hpp>
# include <cppad/local/var_op/div_op.hpp>
# include <cppad/local/var_op/erf_op.hpp>
# include <cppad/local/var_op/exp_op.hpp>
# include <cppad/local/var_op/expm1_op.hpp>
# include <cppad/local/var_op/load_op.hpp>
# include <cppad/local/var_op/log_op.hpp>
# include <cppad/local/var_op/log1p_op.hpp>
# include <cppad/local/var_op/mul_op.hpp>
# include <cppad/local/var_op/neg_op.hpp>
# include <cppad/local/var_op/par_op.hpp>
# include <cppad/local/var_op/pow_op.hpp>
# include <cppad/local/var_op/pri_op.hpp>
# include <cppad/local/var_op/sign_op.hpp>
# include <cppad/local/var_op/sin_op.hpp>
# include <cppad/local/var_op/sinh_op.hpp>
# include <cppad/local/var_op/sqrt_op.hpp>
# include <cppad/local/var_op/sub_op.hpp>
# include <cppad/local/var_op/two_var.hpp>
# include <cppad/local/var_op/one_var.hpp>
# include <cppad/local/var_op/store_op.hpp>
# include <cppad/local/var_op/tan_op.hpp>
# include <cppad/local/var_op/tanh_op.hpp>
# include <cppad/local/var_op/zmul_op.hpp>


# endif
